/* LibMemcached
 * Copyright (C) 2006-2009 Brian Aker, Trond Norbye
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license.  See
 * the COPYING file in the parent directory for full text.
 *
 * Summary: Change the behavior of the memcached connection.
 *
 */

#ifndef __LIBMEMCACHED_CONFIGURE_H
#define __LIBMEMCACHED_CONFIGURE_H

#ifdef __cplusplus
extern "C" {
#endif




#define LIBMEMCACHED_VERSION_STRING "0.47"
#define LIBMEMCACHED_VERSION_HEX 0x00047000

#ifdef __cplusplus
}
#endif

#endif /* __LIBMEMCACHED_CONFIGURE_H */
