// ###########################################
// # DO NOT EDIT! THIS IS A GENERATED FILE 
// ###########################################
    bool isAllowDataLossDuringShutdown() const;
    void setAllowDataLossDuringShutdown(const bool &nval);
    size_t getAlogBlockSize() const;
    void setAlogBlockSize(const size_t &nval);
    std::string getAlogPath() const;
    void setAlogPath(const std::string &nval);
    size_t getAlogSleepTime() const;
    void setAlogSleepTime(const size_t &nval);
    size_t getAlogTaskTime() const;
    void setAlogTaskTime(const size_t &nval);
    std::string getBackend() const;
    void setBackend(const std::string &nval);
    size_t getBgFetchDelay() const;
    void setBgFetchDelay(const size_t &nval);
    size_t getChkMaxItems() const;
    void setChkMaxItems(const size_t &nval);
    size_t getChkPeriod() const;
    void setChkPeriod(const size_t &nval);
    size_t getChkRemoverStime() const;
    void setChkRemoverStime(const size_t &nval);
    bool isConcurrentDB() const;
    void setConcurrentDB(const bool &nval);
    std::string getConfigFile() const;
    void setConfigFile(const std::string &nval);
    std::string getCouchBucket() const;
    void setCouchBucket(const std::string &nval);
    std::string getCouchHost() const;
    void setCouchHost(const std::string &nval);
    size_t getCouchPort() const;
    void setCouchPort(const size_t &nval);
    size_t getCouchReconnectSleeptime() const;
    void setCouchReconnectSleeptime(const size_t &nval);
    size_t getCouchResponseTimeout() const;
    void setCouchResponseTimeout(const size_t &nval);
    bool isDataTrafficEnabled() const;
    void setDataTrafficEnabled(const bool &nval);
    std::string getDbname() const;
    void setDbname(const std::string &nval);
    size_t getExpPagerStime() const;
    void setExpPagerStime(const size_t &nval);
    size_t getExpiryWindow() const;
    void setExpiryWindow(const size_t &nval);
    bool isFailpartialwarmup() const;
    void setFailpartialwarmup(const bool &nval);
    bool isFlushallEnabled() const;
    void setFlushallEnabled(const bool &nval);
    size_t getGetlDefaultTimeout() const;
    void setGetlDefaultTimeout(const size_t &nval);
    size_t getGetlMaxTimeout() const;
    void setGetlMaxTimeout(const size_t &nval);
    size_t getHtLocks() const;
    void setHtLocks(const size_t &nval);
    size_t getHtSize() const;
    void setHtSize(const size_t &nval);
    bool isInconsistentSlaveChk() const;
    void setInconsistentSlaveChk(const bool &nval);
    std::string getInitfile() const;
    void setInitfile(const std::string &nval);
    bool isItemNumBasedNewChk() const;
    void setItemNumBasedNewChk(const bool &nval);
    bool isKeepClosedChks() const;
    void setKeepClosedChks(const bool &nval);
    size_t getKlogBlockSize() const;
    void setKlogBlockSize(const size_t &nval);
    size_t getKlogCompactorQueueCap() const;
    void setKlogCompactorQueueCap(const size_t &nval);
    size_t getKlogCompactorStime() const;
    void setKlogCompactorStime(const size_t &nval);
    std::string getKlogFlush() const;
    void setKlogFlush(const std::string &nval);
    size_t getKlogMaxEntryRatio() const;
    void setKlogMaxEntryRatio(const size_t &nval);
    size_t getKlogMaxLogSize() const;
    void setKlogMaxLogSize(const size_t &nval);
    std::string getKlogPath() const;
    void setKlogPath(const std::string &nval);
    std::string getKlogSync() const;
    void setKlogSync(const std::string &nval);
    size_t getMaxCheckpoints() const;
    void setMaxCheckpoints(const size_t &nval);
    size_t getMaxItemSize() const;
    void setMaxItemSize(const size_t &nval);
    size_t getMaxSize() const;
    void setMaxSize(const size_t &nval);
    size_t getMaxTxnSize() const;
    void setMaxTxnSize(const size_t &nval);
    size_t getMaxVbuckets() const;
    void setMaxVbuckets(const size_t &nval);
    size_t getMemHighWat() const;
    void setMemHighWat(const size_t &nval);
    size_t getMemLowWat() const;
    void setMemLowWat(const size_t &nval);
    size_t getMinDataAge() const;
    void setMinDataAge(const size_t &nval);
    float getMutationMemThreshold() const;
    void setMutationMemThreshold(const float &nval);
    size_t getPagerActiveVbPcnt() const;
    void setPagerActiveVbPcnt(const size_t &nval);
    size_t getPagerUnbiasedPeriod() const;
    void setPagerUnbiasedPeriod(const size_t &nval);
    std::string getPostInitfile() const;
    void setPostInitfile(const std::string &nval);
    size_t getQueueAgeCap() const;
    void setQueueAgeCap(const size_t &nval);
    std::string getStoredValType() const;
    void setStoredValType(const std::string &nval);
    size_t getTapAckGracePeriod() const;
    void setTapAckGracePeriod(const size_t &nval);
    size_t getTapAckInitialSequenceNumber() const;
    void setTapAckInitialSequenceNumber(const size_t &nval);
    size_t getTapAckInterval() const;
    void setTapAckInterval(const size_t &nval);
    size_t getTapAckWindowSize() const;
    void setTapAckWindowSize(const size_t &nval);
    float getTapBackfillResident() const;
    void setTapBackfillResident(const float &nval);
    size_t getTapBacklogLimit() const;
    void setTapBacklogLimit(const size_t &nval);
    float getTapBackoffPeriod() const;
    void setTapBackoffPeriod(const float &nval);
    size_t getTapBgMaxPending() const;
    void setTapBgMaxPending(const size_t &nval);
    size_t getTapKeepalive() const;
    void setTapKeepalive(const size_t &nval);
    size_t getTapNoopInterval() const;
    void setTapNoopInterval(const size_t &nval);
    float getTapRequeueSleepTime() const;
    void setTapRequeueSleepTime(const float &nval);
    size_t getTapThrottleCapPcnt() const;
    void setTapThrottleCapPcnt(const size_t &nval);
    ssize_t getTapThrottleQueueCap() const;
    void setTapThrottleQueueCap(const ssize_t &nval);
    size_t getTapThrottleThreshold() const;
    void setTapThrottleThreshold(const size_t &nval);
    bool isVb0() const;
    void setVb0(const bool &nval);
    bool isWaitforwarmup() const;
    void setWaitforwarmup(const bool &nval);
    bool isWarmup() const;
    void setWarmup(const bool &nval);
    size_t getWarmupBatchSize() const;
    void setWarmupBatchSize(const size_t &nval);
    size_t getWarmupMinMemoryThreshold() const;
    void setWarmupMinMemoryThreshold(const size_t &nval);
    size_t getWarmupMinItemsThreshold() const;
    void setWarmupMinItemsThreshold(const size_t &nval);
