/* LibMemcached
 * Copyright (C) 2006-2009 Brian Aker
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license.  See
 * the COPYING file in the parent directory for full text.
 *
 * Summary: Internal functions used by the library. Not for public use!
 *
 */

#ifndef __LIBMEMCACHED_INTERNAL_H__
#define __LIBMEMCACHED_INTERNAL_H__

#if defined(BUILDING_LIBMEMCACHED)

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus
}
#endif

#endif /* BUILDING_LIBMEMCACHED */
#endif /* __LIBMEMCACHED_INTERNAL_H__ */
