#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class Reader(object):
    def __init__(self, fp):
        pass

    def __iter__(self):
        return self

    def next(self):
        raise NotImplementedError

    def close(self):
        pass


class Writer(object):
    def __init__(self, fp):
        pass

    def write(self, record):
        raise NotImplementedError

    def close(self):
        pass
