/* LibMemcached
 * Copyright (C) 2006-2009 Brian Aker
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license.  See
 * the COPYING file in the parent directory for full text.
 *
 * Summary: Connection pool library.
 *
 * Author: Trond Norbye, Brian Aker
 *
 */


#ifndef __LIBMEMCACHED__MEMCACHED_UTIL_H__
#define __LIBMEMCACHED__MEMCACHED_UTIL_H__

#include <libmemcached/util/ping.h>
#include <libmemcached/util/pool.h>
#include <libmemcached/util/version.h>

#endif /* __LIBMEMCACHED__MEMCACHED_UTIL_H__ */
