var AllImages = ["images/server_icon.gif", "images/usages_new.png", "images/warning_bg.gif", "images/usage_small_2.gif", "images/server_status_icons_solid_bg.png", "images/config-bottom.png", "images/alert_green.png", "images/button_gray_big.gif", "images/add_back_btn.png", "images/ic_plus.png", "images/large_buttons_bg.gif", "images/usage_small.gif", "images/header_bg.png", "images/active_bg.png", "images/expander_bg.gif", "images/button_gray_medium.gif", "images/auth_form_bg.png", "images/exp_filter2.png", "images/button_blue.gif", "images/background.png", "images/ico_server_big.png", "images/button_gray_200.gif", "images/devider_line_2.gif", "images/config-top.png", "images/server_status_icons.png", "images/active_hover_bg.png", "images/usages_third.png", "images/bg_key_inp_text.png", "images/spinner/spinner_gr_nav.gif", "images/spinner/spinner2_gr_nav.gif", "images/spinner/spinner_lightbox.gif", "images/spinner/spinner_3.gif", "images/spinner/spinner_1.gif", "images/usages_two.png", "images/blue-button.png", "images/bg_li_th.gif", "images/edit-button.png", "images/sign_in_button.gif", "images/server_icon.png", "images/welcome-setup-button.png", "images/attention_bg.gif", "images/buttons.png", "images/info_bg.gif", "images/couchbase_logo.png", "images/mini_graph_bg.gif", "images/add_btn.png", "images/btn_1_bg.png", "images/crn_hdr_err.gif", "images/tooltip_bg.gif", "images/usages.png", "images/switcher_bg.png", "images/alert_red.png", "images/tooltip_bg_777777.png", "images/devider_line_1.gif", "images/add_back_bg.png", "images/exp_filter.png", "images/btn_2_bg.png", "images/ic_prev_next_arr.png", "images/warning_bg_2.png", "images/logo_large.png", "images/intro_logo.png", "images/bg_h2.gif", "images/footer_bg.png", "images/navigation_bg.png", "images/icons.png", "images/home_icon.png", "images/button_red.gif", "images/button_blue_3.gif", "images/expander_2_bg.gif", "images/btn_x.png", "images/bg_total.gif", "images/ui-theme/ui-icons_2e83ff_256x240.png", "images/ui-theme/ui-icons_cd0a0a_256x240.png", "images/ui-theme/ui-bg_flat_75_ffffff_40x100.png", "images/ui-theme/ui-bg_glass_75_dadada_1x400.png", "images/ui-theme/ui-bg_glass_55_fbf9ee_1x400.png", "images/ui-theme/ui-icons_888888_256x240.png", "images/ui-theme/ui-bg_glass_95_fef1ec_1x400.png", "images/ui-theme/ui-bg_glass_65_ffffff_1x400.png", "images/ui-theme/ui-icons_222222_256x240.png", "images/ui-theme/ui-bg_glass_75_e6e6e6_1x400.png", "images/ui-theme/ui-icons_454545_256x240.png", "images/ui-theme/ui-bg_highlight-soft_75_cccccc_1x100.png"];
