/////////////////////////////////
// Generated file, do not edit //
/////////////////////////////////
#ifndef GENERATED_23866_1355147971
#define GENERATED_23866_1355147971

#ifdef __cplusplus
extern "C" {
#endif

const char *get_stats_info(void);

#ifdef __cplusplus
}
#endif
#endif
